using System.Numerics;
using Frontend.Functions;
using Frontend.Events;
using Nethereum.Contracts;
using Nethereum.Hex.HexTypes;
using Nethereum.Web3;
using Nethereum.Web3.Accounts;

using Frontend.Configuration;

namespace Frontend;

public class PoolV3Client
{
    private readonly Contract _pool;
    private readonly Web3 _web3;
    private readonly ChainSettings _chainSettings;

    private string? token0Address = null;
    private string? token1Address = null;
    private uint? poolFeeTier = null;

    public PoolV3Client(Web3 web3, ChainSettings chainSettings)
    {

        _pool = web3.Eth.GetContract(
            File.ReadAllText(
                Path.Combine(
                Directory.GetCurrentDirectory(),
                    "Configuration",
                    "Abi",
                    "uniswapV3Pool.abi.json"
                )
            ),
            chainSettings.Uniswap.RocEthPoolAddress
        );
        _web3 = web3;
        _chainSettings = chainSettings;
    }

    public async Task<decimal> GetPairRatio()
    {
        var handler = _web3.Eth.GetContractQueryHandler<Slot0Funtion>();
        var func = new Slot0Funtion();
        var slot0Result = await handler.QueryDeserializingToObjectAsync<Slot0OutputDTO>(func, _pool.Address);
        var sqrtPriceX96 = slot0Result.SqrtPriceX96;
        return (decimal)SqrtPriceX96ToPrice(sqrtPriceX96);
    }

    public double SqrtPriceX96ToPrice(BigInteger sqrtPriceX96)
    {
        var priceX96 = sqrtPriceX96 * sqrtPriceX96;
        return (double)priceX96 / Math.Pow(2, 192);
    }

    public BigInteger PriceToSqrtPriceX96(double price)
    {
        return new BigInteger(Math.Sqrt(price) * Math.Pow(2, 96));
    }
    public async Task<string> Token0Address()
    {
        if (null == token0Address)
        {
            var handler = _web3.Eth.GetContractQueryHandler<Token0Function>();
            var func = new Token0Function();
            var result = await handler.QueryAsync<string>(_pool.Address, func);
            token0Address = result;
        }

        return token0Address;
    }

    public async Task<string> Token1Address()
    {
        if (null == token1Address)
        {
            var handler = _web3.Eth.GetContractQueryHandler<Token1Function>();
            var func = new Token1Function();
            var result = await handler.QueryAsync<string>(_pool.Address, func);
            token1Address = result;
        }

        return token1Address;
    }

    public async Task<uint> getPoolFeeTier()
    {
        if (null == poolFeeTier)
        {
            var handler = _web3.Eth.GetContractQueryHandler<FeeFunction>();
            var func = new FeeFunction();
            var result = await handler.QueryAsync<int>(_pool.Address, func);
            poolFeeTier = (uint)result;
        }
        return (uint)poolFeeTier;
    }


    /*
    this function should swap the tokens in the pool
    */
    public async Task<decimal> SwapAsync(string fromTokenAddress, string toTokenAddress, decimal amount, Account account, decimal amountOutMinimum=0, uint sqrtPriceLimitX96=0)
    {
        var swapFunction = new SwapExactInputSingleFunction
        {
            TokenIn = fromTokenAddress,
            TokenOut = toTokenAddress,
            AmountIn = Web3.Convert.ToWei(amount),
            AmountOutMinimum = Web3.Convert.ToWei(amountOutMinimum), // Set your minimum amount out
            Recipient = account.Address,
            Fee = new BigInteger(await getPoolFeeTier()),
            SqrtPriceLimitX96 = sqrtPriceLimitX96
        };

        swapFunction.Gas= new BigInteger(300000);

        var handler = _web3.Eth.GetContractTransactionHandler<SwapExactInputSingleFunction>();
       
        var transactionReceipt = await handler.SendRequestAndWaitForReceiptAsync(_chainSettings.Uniswap.SwapRouterV2Address, swapFunction);
        var swapEvent = transactionReceipt.DecodeAllEvents<SwapEventDTO>();
        if (swapEvent.Count > 0)
        {
            var receivedAmount = swapEvent[0].Event.AmountOut;
            return Web3.Convert.FromWei(receivedAmount);
        }


        throw new Exception("Swap event not found in transaction receipt");
    }


}