using System.Numerics;
using Frontend.Functions;
using Nethereum.Contracts;
using Nethereum.Contracts.Standards.ERC20.ContractDefinition;
using Nethereum.Web3;
using Nethereum.Web3.Accounts;

namespace Frontend;

public class TokenClient
{
    const int Decimals = 18;

    private readonly Account _account;
    private readonly Web3 _web3Client;
    private readonly string _tokenAddress;
    private readonly string _tokenAbi;

    private Contract Contract 
    {
        get => _web3Client.Eth.GetContract(_tokenAbi, _tokenAddress);
    }

    public TokenClient(
        string apiUrl,
        string tokenAddress,
        string tokenAbi,
        string accountPrivateKey,
        int chainId
    ) {
        _tokenAddress = tokenAddress;
        _tokenAbi = tokenAbi;
        _account = new(accountPrivateKey, chainId: chainId);
        _web3Client = new(_account, apiUrl);
    }

    public async Task<decimal> TotalSupply()
    {
        var totalSupplyHandler = _web3Client.Eth.GetContractQueryHandler<TotalSupplyFunction>();
        var totalSupplyFunction = new TotalSupplyFunction();
        var totalSupply = await totalSupplyHandler.QueryAsync<BigInteger>(Contract.Address, totalSupplyFunction);
        return ConvertSmallUnitToBigUnit(totalSupply);
    }

    public async Task Mint(string recipientAddress, BigInteger amount, CancellationToken cancellationToken)
    {
        var mintHandler = _web3Client.Eth.GetContractTransactionHandler<MintFunction>();
        var mintFunction = new MintFunction()
        {
            To = recipientAddress,
            Amount = amount,
        };

        await mintHandler.SendRequestAndWaitForReceiptAsync(Contract.Address, mintFunction, cancellationToken);
    }

    public async Task Burn(BigInteger amount, CancellationToken cancellationToken)
    {
        var burnHandler = _web3Client.Eth.GetContractTransactionHandler<BurnFunction>();
        var burnFunction = new BurnFunction()
        {
            Amount = amount,
        };

        await burnHandler.SendRequestAndWaitForReceiptAsync(Contract.Address, burnFunction, cancellationToken);
    }

    public decimal ConvertSmallUnitToBigUnit(BigInteger smallUnit)
    {
        return (decimal)(smallUnit / (BigInteger)Math.Pow(10, Decimals));
    }
}
